//------------------------------------------------------------------------------
//
//	npmouse.c
//
//  ANTz application based on NPE is free software
//	NPE is the Neural Physics Engine that drives the interactions
//
//	ANTz hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
//	Copyright (c) 2010-2011, Shane Saxon
//	All rights reserved.
//	Redistribution and use in source and binary forms, with or without 
//	modification, are permitted provided that the following conditions are met:
//	-	Redistributions of source code must retain the above copyright notice, 
//		this list of conditions and the following disclaimer.
//	-	Redistributions in binary form must reproduce the above copyright 
//		notice, this list of conditions and the following disclaimer in the 
//		documentation and/or other materials provided with the distribution.
//	-	Neither the name of the Shane Saxon nor the names of its contributors 
//		may be used to endorse or promote products derived from this software 
//		without specific prior written permission.
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
//	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
//	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
//	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
//	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
//	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
//	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
//	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
//	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
//	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
//	POSSIBILITY OF SUCH DAMAGE.
//
//------------------------------------------------------------------------------

#include "npmouse.h"

#include "../npctrl.h"
#include "../npdata.h"

#include "npgl.h"

#include "watchtools.h"


//MB-CAMERA
#define EXAMINERCAMERA	1
#define FLYERCAMERA		2

#define NOBUTTON -1
#define BOTH_BUTTONS 3
//MB-END
 
float lowPassFilter( float start, float target, float fac );

//------------------------------------------------------------------------------
void npInitMouse (void* dataRef)
{
	pData data = (pData) dataRef;

	//zz-CAMERA
	//compare intialization values with working project, debug zz

	data->io.mouse.phi		= kPI/2.0f;
	data->io.mouse.theta	= 0.0f;
	data->io.mouse.zeta		= 0.0f;

	data->io.mouse.ePhi		= kPI/2.0f;
	data->io.mouse.eTheta	= 0.0f;

	data->io.mouse.cameraMode = FLYERCAMERA;
	data->io.mouse.tiltCamera = false;

	data->io.mouse.target.x = 0.0f;
	data->io.mouse.target.y = 0.0f;
	data->io.mouse.target.z = 0.0f;

	data->io.mouse.targetDest.x = 0.0f;
	data->io.mouse.targetDest.y = 0.0f;
	data->io.mouse.targetDest.z = 0.0f;

	data->io.mouse.targetRadius = 0.0f;
	data->io.mouse.targetRadiusDest = 0.0f;

	data->io.mouse.translating = false;

	//zz-END
//------------------------------------------

	data->io.mouse.x = 0;						//raw coordinate
	data->io.mouse.y = 0;
	data->io.mouse.z = 0;						//typically the scroll wheel

	data->io.mouse.mode = kNPmouseModeNull;

	data->io.mouse.buttonL = false;					//true when pressed
	data->io.mouse.buttonC = false;
	data->io.mouse.buttonR = false;

	data->io.mouse.screenX = 0;					//screen position in pixels
	data->io.mouse.screenY = 0;

	data->io.mouse.previousX = 0;
	data->io.mouse.previousY = 0;
	data->io.mouse.previousZ = 0;

	data->io.mouse.theta = 0.0f;
	data->io.mouse.zeta = 0.0f;

	data->io.mouse.cmDX = 0.0f;
	data->io.mouse.cmDY = 0.0f;

	data->io.mouse.deltaX = 0.0f;					//mouse vector
	data->io.mouse.deltaY = 0.0f;					//clears every mouse event
	data->io.mouse.deltaZ = 0.0f;

	data->io.mouse.accumulateX = 0;			//clears each NPE period
	data->io.mouse.accumulateY = 0;
	data->io.mouse.accumulateZ = 0;

	data->io.mouse.target.x = 0.0f;
	data->io.mouse.target.y = 0.0f;
	data->io.mouse.target.z = 0.0f;

	data->io.mouse.examinerCenter.x = 0.0f;
	data->io.mouse.examinerCenter.y = 0.0f;
	data->io.mouse.examinerCenter.z = 0.0f;

	data->io.mouse.examinerCenterDest.x = 0.0f;
	data->io.mouse.examinerCenterDest.y = 0.0f;
	data->io.mouse.examinerCenterDest.z = 0.0f;

	data->io.mouse.examinerRadiusDest = 0.0f;
	data->io.mouse.examinerRadius = 0.0f;
	data->io.mouse.radiusFac = 0.0f;

	data->io.mouse.pickedOnThisFrame = false;
	data->io.mouse.pinSelected = false;
	
	data->io.mouse.size = sizeof(NPmouse);
}


//------------------------------------------------------------------------------
void npMouseUpdate (int x, int y)
{
	float dX = 0;
	float dY = 0;
	static float tX = 0.0f;
	static float tY = 0.0f;
	
	pData data = (pData) npGetDataRef();

	if (data->map.nodeRootCount > kNPpinRoot)
	{
		npCtrlCommand (kGraphSelectAllOff, data);	//zz debug not efficient
		npCtrlCommand (kGraphCamera, data);
	}

	dX = data->io.mouse.deltaX;
	dY = data->io.mouse.deltaY;

	//Examiner
	if( data->io.mouse.mode == kNPmouseModeCamExamXY
		|| data->io.mouse.mode == kNPmouseModeCamExamXZ)		//MB-CAMERA
	{
		tX = (float)dX / 300.0f;
		tY = (float)dY / 300.0f;
		data->io.mouse.eTheta += tX;

		if( data->io.mouse.mode == kNPmouseModeCamExamXY )
		{
			data->io.mouse.ePhi += -tY;
			if( data->io.mouse.ePhi < 0.01f )		//Close to Zero
				data->io.mouse.ePhi = 0.01f;
			if( data->io.mouse.ePhi > kPI/2.0f )
				data->io.mouse.ePhi = kPI/2.0f;
		}
		else
			data->io.mouse.radiusFac *= 1.0f-dY/500.0f;
	}
	//Flyer
	if( data->io.mouse.mode == kNPmouseModeCamFlyA )
	{
		data->io.mouse.cmDX += (float)dX / 1000.0f;
		data->io.mouse.cmDY += (float)dY / 1000.0f;
	}

}


//accumulates and stores the mouse delta coordinates
//------------------------------------------------------------------------------
void npMouseMotion (int x, int y)
{
	pData data = (pData) npGetDataRef();

	//accumulate mouse delta coords
	data->io.mouse.deltaX += (GLfloat)(x - data->io.mouse.previousX);
	data->io.mouse.deltaY += (GLfloat)(y - data->io.mouse.previousY);

	data->io.mouse.previousX = x;
	data->io.mouse.previousY = y;

//	if( !data->io.picking )
//		npMouseUpdate (x,y);
}

void npSetCamTarget(void)
{
	float dX = 0.0f, dY = 0.0f, dZ = 0.0f;
	float cdX = 0.0f, cdY = 0.0f, cdZ = 0.0f;
	float tX = 0.0f, tY = 0.0f, tZ = 0.0f;

	pData data = (pData) npGetDataRef();

	NPnodePtr cam = cameraSelected();
	NPnodePtr node = data->map.node[data->map.selectedIndex];

	if (node == NULL)
		return;

	data->io.mouse.targetDest.x = node->translate.x;
	data->io.mouse.targetDest.y = node->translate.y;
	data->io.mouse.targetDest.z = node->translate.z + 5.0f * node->scale.z;

	dX = cam->translate.x - data->io.mouse.targetDest.x;
	dY = cam->translate.y - data->io.mouse.targetDest.y;
	dZ = cam->translate.z - data->io.mouse.targetDest.z;

	data->io.mouse.targetRadius = sqrtf(dX*dX + dY*dY + dZ*dZ);

	//true on first pass
	if (cam->proximityRadius.x == 0.0f)
		cam->proximityRadius.x = data->io.mouse.targetRadius;

	//set target based on radius
	cdX = -sinf(cam->rotateRad.y) * cam->proximityRadius.x;
	cdY = -cosf(cam->rotateRad.y) * cam->proximityRadius.x;
	cdZ = sinf(cam->rotateRad.x) * cam->proximityRadius.x;

	tX = cdX + data->io.mouse.targetDest.x;
	tY = cdY + data->io.mouse.targetDest.y;
	tZ = cdZ + data->io.mouse.targetDest.z;

	data->io.mouse.target.x = tX;
	data->io.mouse.target.y = tY;
	data->io.mouse.target.z = tZ;

	data->io.mouse.translating = true;
	
	printf ("cdX: %f  %f  %f \n", cdX,cdY,cdZ);
	printf ("tX: %f  %f  %f \n", tX,tY,tZ);

	printf ("%f  %f  %f  r: %f  rD: %f\n", data->io.mouse.targetDest.x,
							data->io.mouse.targetDest.y,
							data->io.mouse.targetDest.z,
							data->io.mouse.targetRadius,
							cam->proximityRadius.x);

}

//------------------------------------------------------------------------------
void npMouseEvent (int button, int state, int x, int y)
{
	pData data = (pData) npGetDataRef();

	NPnodePtr node = NULL;

							//causes objects to jump on click deltaX, debug zz
//	data->io.mouse.deltaX = (float)(x - data->io.mouse.previousX);
//	data->io.mouse.deltaY = (float)(y - data->io.mouse.previousY);

	//	glutWarpPointer(150,150);	//debug, zz

	data->io.mouse.previousX = x;
	data->io.mouse.previousY = y;	
	data->io.mouse.deltaX = 0.0f;
	data->io.mouse.deltaY = 0.0f;
	data->io.mouse.cmDX = 0.0f;
	data->io.mouse.cmDY = 0.0f;


	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN)
		data->io.mouse.buttonL = true;
	else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP)
		data->io.mouse.buttonL = false;

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
		data->io.mouse.buttonR = false;
	else if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN)
		data->io.mouse.buttonR = true;

	if( state == GLUT_DOWN )	//Mouse Click Down (Press)
	{
		PickPass( true );
		npGLDrawScene( data );		//optimize to draw only picked region, debug zz
		y = height - y;
		npPick( x, y );
		npGLDrawScene( data );							//zz needed to select object... debug
		PickPass( false );

		// set the examiner target on the selected node
		// for use if examiner mode entered next
		npSetCamTarget();
	}

	if( state == GLUT_UP )						//Mouse Click Up (Release)
	{	
		data->io.picking = false;
		PickPass( false );
	}

	//calculate mouse operation mode
	if (data->io.picking)				//picked an object
	{
		if ( data->io.mouse.buttonL && !data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeDragXY;
		if ( !data->io.mouse.buttonL && data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeCamExamXZ; // kNPmouseModeDragXZ;
		if ( data->io.mouse.buttonL && data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeCamExamXY; //kNPmouseModeDragWorld;
	}
	else								//picked background
	{
		if ( data->io.mouse.buttonL && !data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeCamLook;//kNPmouseModeCamExamXZ;
		if ( !data->io.mouse.buttonL && data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeCamFlyA;
		if ( data->io.mouse.buttonL && data->io.mouse.buttonR )
			data->io.mouse.mode = kNPmouseModeCamExamXY;

		// select the camera if not selected and button down
		if (data->map.selectedIndex > kNPgridRoot && state == GLUT_DOWN )
		{
			npCtrlCommand (kGraphSelectAllOff, data);	//zz debug unselects all, fix this
			npCtrlCommand (kGraphCamera, data);
		}
	}

	// if no buttons
	if ( !data->io.mouse.buttonL && !data->io.mouse.buttonR )
		data->io.mouse.mode = kNPmouseModeNull;

	printf(" mode: %d \n", data->io.mouse.mode);		//remove, debug, zz
}


//------------------------------------------------------------------------------
float lowPassFilter( float start, float target, float fac )
{
	return (float)(fac*target + (1.0f-fac)*start);
}


//------------------------------------------------------------------------------
NPfloatXYZ lookVec()
{
	pData data = (pData) npGetDataRef();
	NPnodePtr camNode = cameraSelected();
	
	NPfloatXYZ look;
	NPfloatXYZ eCent = data->io.mouse.examinerCenter; //examinerCenterDest;
	
	float dX = eCent.x - camNode->translate.x;
	float dY = eCent.y - camNode->translate.y;
	float dZ = eCent.z - camNode->translate.z;

	//normalize
	float len = (float)sqrt( dX*dX + dY*dY + dZ*dZ );
	dX /= len;
	dY /= len;
	dZ /= len;
	look.x = dX;
	look.y = dY;
	look.z = dZ;
	
	return look;
}


void npExaminerCamera( void* dataRef )
{	
	float fac = 0.04f;
	float radius = 0.0f;
	float tX = 0.0f, tY = 0.0f, tZ = 0.0f;
	float rX = 0.0f, rY = 0.0f, rZ = 0.0f;
	float aX = 0.0f, aY = 0.0f, aZ = 0.0f;

	pData data = (pData) dataRef;

	NPnodePtr camNode = cameraSelected();	
	
	//Move to Center over time
	data->io.mouse.examinerCenter.x = lowPassFilter( data->io.mouse.examinerCenter.x, data->io.mouse.examinerCenterDest.x, fac );
	data->io.mouse.examinerCenter.y = lowPassFilter( data->io.mouse.examinerCenter.y, data->io.mouse.examinerCenterDest.y, fac );
	data->io.mouse.examinerCenter.z = lowPassFilter( data->io.mouse.examinerCenter.z, data->io.mouse.examinerCenterDest.z, fac );
	data->io.mouse.examinerRadius   = lowPassFilter( data->io.mouse.examinerRadius,   data->io.mouse.examinerRadiusDest,   fac );

	radius = data->io.mouse.examinerRadius;
	
	aX = data->io.mouse.examinerCenter.x;
	aY = data->io.mouse.examinerCenter.y;
	aZ = data->io.mouse.examinerCenter.z;
	
	radius *= data->io.mouse.radiusFac;
	if( radius <= 0.01f )		//CLAMP "Close"
		radius  = 0.01f;
	
	if( data->io.mouse.ePhi < 0.01f )		//Close to Zero
		data->io.mouse.ePhi = 0.01f;
	if( data->io.mouse.ePhi > kPI/2.0f )
		data->io.mouse.ePhi = lowPassFilter( data->io.mouse.ePhi, kPI/2.0f, 0.03f);
	
	//Spherical to Cartesian
	tX = aX + radius*sinf(data->io.mouse.ePhi)*sinf(data->io.mouse.eTheta);
	tY = aY + radius*sinf(data->io.mouse.ePhi)*cosf(data->io.mouse.eTheta);
	tZ = aZ + radius*cosf(data->io.mouse.ePhi);
	
	rX = aX - tX;
	rY = aY - tY;
	rZ = aZ - tZ;
	
	camNode->translate.x = tX;
	camNode->translate.y = tY;
	camNode->translate.z = tZ;
	camNode->rotate.x	 = rX;
	camNode->rotate.y	 = rY;
	camNode->rotate.z	 = rZ;
}


NPfloatXYZA lastFlyerRotate;
void npFlyerCamera( void* dataRef )
{
	pData data = (pData) dataRef;

	NPnodePtr camNode = cameraSelected();
	
	data->io.mouse.theta += data->io.mouse.cmDX/20.0f;
	if( data->io.mouse.tiltCamera )
		data->io.mouse.phi += data->io.mouse.cmDX/20.0f;

	if( data->io.mouse.mode == kNPmouseModeNull )
	{
		data->io.mouse.cmDX *= 0.9f;
		data->io.mouse.cmDY *= 0.9f;
	}
	if( data->io.mouse.buttonR)
	{
		data->io.mouse.cmDX *= 1.0f;
		data->io.mouse.cmDY *= 1.0f;
	}
	
//	WATCH_FLOAT( data->io.mouse.theta );
//	camNode->rotate.x = sinf(data->io.mouse.phi)*sinf(data->io.mouse.theta);
//	camNode->rotate.y = sinf(data->io.mouse.phi)*cosf(data->io.mouse.theta);
//	camNode->rotate.z = cosf(data->io.mouse.phi);
//	camNode->rotateRad.x += data->io.mouse.cmDY/20.0f;
	camNode->rotateRad.y += data->io.mouse.cmDX/20.0f;


	//Flatten out Phi
//	data->io.mouse.phi = lowPassFilter( data->io.mouse.phi, kPI/2.0, 0.01 );
	
	lastFlyerRotate = camNode->rotate;
//	WATCH_VEC( lastFlyerRotate );
	
	if( !data->io.mouse.tiltCamera )
	{
		camNode->translate.x += -data->io.mouse.cmDY*camNode->rotate.x;
		camNode->translate.y += -data->io.mouse.cmDY*camNode->rotate.y; 
		camNode->translate.z += -data->io.mouse.cmDY*camNode->rotate.z; 
	}
}


float angleBetweenVectors( float aX,float aY,  float bX, float bY )
{
	//	atan2(v2.y,v2.x) - atan2(v1.y,v1.x)
	float angle = atan2f( bY,bX ) - atan2f( aY,aX );
	return angle;
} 

void cartToSphere( float x, float y, float z, float* th, float* ph )
{
	float r;

	WATCH_FLOAT( x );
	WATCH_FLOAT( y );
	WATCH_FLOAT( z );
	WATCH_FLOAT( *th );
	WATCH_FLOAT( *ph );
	

	r = sqrtf( x*x + y*y + z*z );
	
//	*ph = acosf(  z/r );
	*ph = atan2f( sqrtf( x*x + y*y ), z );
	*th = atan2f( x,y );						//X & Y Swap
//	*th = atan2f( y,x );
}

void transitionExaminerToFlyer()
{		
	float th, ph;
	NPfloatXYZ look;

	pData data = (pData) npGetDataRef();

	WATCH_FUNC;
//	data->io.mouse.cmDY = 0.0;
//	data->io.mouse.cmDX = 0.0;
	
//	NPnodePtr camNode = cameraSelected();
	

	look = lookVec();
/*	
	float th;
	th = -1.0 * angleBetweenVectors( 0.0,1.0,  look.x,look.y );
	data->io.mouse.theta = th;
//	WATCH_FLOAT( data->io.mouse.theta );
*/

	cartToSphere( look.x,look.y,look.z, &th,&ph );
	data->io.mouse.theta = th;
	data->io.mouse.phi   = ph;
	
	WATCH_FLOAT( data->io.mouse.theta );
	WATCH_FLOAT( data->io.mouse.phi   );
}

void transitionFlyerToExaminer()
{
	float th, ph;
	float rad;
	NPnodePtr camNode  = NULL;
	NPfloatXYZ trans;
	NPfloatXYZA rotate;

	pData data = (pData) npGetDataRef();

	WATCH_FUNC;
	
	rad = data->io.mouse.examinerRadius;
	camNode  = cameraSelected();
	trans   = camNode->translate;
	rotate = lastFlyerRotate;
		
	WATCH_VEC( rotate );

	//	examinerCenterDest = trans + examinerRadius*rotate;
	data->io.mouse.examinerCenterDest.x = trans.x + rad*rotate.x;
	data->io.mouse.examinerCenterDest.y = trans.y + rad*rotate.y;
	data->io.mouse.examinerCenterDest.z = trans.z + rad*rotate.z;
	
	data->io.mouse.examinerCenter = data->io.mouse.examinerCenterDest;
	data->io.mouse.examinerRadius = data->io.mouse.examinerRadiusDest;
	//	data->io.mouse.examinerRadius *= data->io.mouse.radiusFac;
	data->io.mouse.radiusFac = 1.0f;
	
//	data->io.mouse.eTheta = -1.0 * angleBetweenVectors( 0.0,1.0, rotate.x,rotate.y );
//	data->io.mouse.eTheta -= kPI;
	

	cartToSphere( rotate.x,rotate.y,rotate.z, &th,&ph );
	th += kPI;
	ph = kPI - ph;
	data->io.mouse.eTheta = th;
	data->io.mouse.ePhi   = ph;
	
	WATCH_FLOAT( data->io.mouse.eTheta );
}

//MB-CAMERA
void npCameraUpdate( void* dataRef )
{
	static int lastCameraMode = FLYERCAMERA;
	
	pData data = (pData) dataRef;

	if( data->io.mouse.buttonL )
		data->io.mouse.cameraMode = EXAMINERCAMERA;
	else if( data->io.mouse.buttonR )
		data->io.mouse.cameraMode = FLYERCAMERA;

	if( data->io.mouse.buttonL && data->io.mouse.buttonR )//data->io.mouse.mode == kNPmouseModeLR )
		data->io.mouse.tiltCamera = true;
	else 
		data->io.mouse.tiltCamera = false;
	 
	//handle camera transition
	if (data->io.mouse.cameraMode != lastCameraMode)
	{
		if( data->io.mouse.cameraMode == FLYERCAMERA )
			transitionExaminerToFlyer();
		if( data->io.mouse.cameraMode == EXAMINERCAMERA )
			transitionFlyerToExaminer();
	}
	
	//update cameras
	if( data->io.mouse.cameraMode == EXAMINERCAMERA )
			npExaminerCamera( dataRef );
	if( data->io.mouse.cameraMode == FLYERCAMERA )
			npFlyerCamera( dataRef );
	

	lastCameraMode = data->io.mouse.cameraMode;
}

